//---------------------------------------------------------------------------
#include <vcl.h>
#include <io.h>
#include <stdlib.h>
#pragma hdrstop

#include "DataStorage.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
__fastcall TFileStorage::TFileStorage(System::AnsiString asFile, TNotifyEvent FOnInitCacheBlock,
                                        TFOnInit FOnInit):TObject()
{
 int i, Rd = 0, EsimatedBlockCount;
 FileHeader = new TStorageHeader;
 FlushList = new TList();
 CacheList = new TList();
 DelList = new TList();
 FirstList = new TList();
 TagList = new TList();

 TRecHeader *BlockHdr, *CacheRec;
 TTagHeader *TagHeader;
 DataSamples = 0;
 CacheCount = 10;
 if (access(asFile.c_str(), 0) == -1)
 {
  DataFile = new TFileStream(asFile, fmCreate|fmShareExclusive);
  FileHeader->RecSize = BUF_SIZE;
  FileHeader->MaxFileSize = 12800;
  DataFile->Write(FileHeader, sizeof(TStorageHeader));
  for (i = 0; i < MAX_DID; i++) FileHeader->LastTag[i] = -1;
 }
 else
 {
  DataFile = new TFileStream(asFile, fmOpenReadWrite|fmShareExclusive);
  DataFile->Read(FileHeader, sizeof(TStorageHeader));
 }

 for ( i = 0; i < MAX_DID; i++)
 {
  DIDs[i] = new TDIDState();
  DIDs[i]->LastTag = -1;
  DIDs[i]->Marker = 'X';
  DIDs[i]->LastIndex = -1;
  DIDs[i]->BufSize = 0;
  DIDs[i]->Bytes = 0;
 }
 EsimatedBlockCount = (DataFile->Size - sizeof(TStorageHeader)) / (BUF_SIZE*CacheCount);
 if (FOnInit) FOnInit(EsimatedBlockCount);
 while (DataFile->Position < DataFile->Size)
 {
  Rd = DataFile->Read(DIDs[0]->Data, FileHeader->RecSize);
  if (Rd < FileHeader->RecSize) break;
  if ( ((TRecHeader *)&DIDs[0]->Data[0])->Marker != 'C' ) continue;
  BlockHdr = new TRecHeader;
  memcpy(BlockHdr, &DIDs[0]->Data[0], sizeof(TRecHeader));
  AddCacheRec(BlockHdr, 0);
  for (i = 1; i < BlockHdr->Index; i++)
  {
   CacheRec = (TRecHeader *)&DIDs[0]->Data[i*sizeof(TRecHeader)];
   switch (CacheRec->Marker) {
    case 'T':
     TagHeader = new TTagHeader;
     DataFile->Seek(CacheRec->DataOffset, soFromBeginning);
     DataFile->Read(TagHeader, sizeof(TTagHeader));
     TagList->Add(TagHeader);
    case 'S':
    case 'F':
    case 'C':
     CacheRec = new TRecHeader;
     memcpy(CacheRec, &DIDs[0]->Data[i*sizeof(TRecHeader)], sizeof(TRecHeader));
     AddCacheRec(CacheRec, 0);
    break;
   }
  }
  if (FOnInitCacheBlock) FOnInitCacheBlock(this);
  DataFile->Seek(BlockHdr->DataOffset, soFromBeginning);
  DataFile->Seek(BlockHdr->Index*FileHeader->RecSize, soFromCurrent);
 }
}
//---------------------------------------------------------------------------
__fastcall TFileStorage::~TFileStorage()
{
 int i;
 FlushCache(FM_FULL);
 for (i = 0; i < MAX_DID; i++) delete DIDs[i];

 DataFile->Seek(0, soFromBeginning);
 DataFile->Write(FileHeader, sizeof(TStorageHeader));

 for (i = 0; i < TagList->Count; i++)
  delete (TTagHeader *)TagList->Items[i];
 for (i = 0; i < FlushList->Count; i++)
  delete (TRecHeader *)FlushList->Items[i];
 delete FlushList;
 for (i = 0; i < CacheList->Count; i++)
  delete (TRecHeader *)CacheList->Items[i];
 delete CacheList;
 for (i = 0; i < DelList->Count; i++)
  delete (TRecHeader *)DelList->Items[i];
 delete DelList;
 for (i = 0; i < FirstList->Count; i++)
  delete (TRecHeader *)FirstList->Items[i];
 delete FirstList;
 delete FileHeader;
 delete DataFile;
}
//---------------------------------------------------------------------------
TRecHeader* __fastcall TFileStorage::FindFreeBlock()
{
 TRecHeader *CacheRec;
 int i;
 for (i = 0; i < DelList->Count; i++)
 {
  CacheRec = (TRecHeader *)DelList->Items[i];
  if (CacheRec->Marker == 'X') return CacheRec;
 }
 if (DataSamples < FileHeader->MaxFileSize) return NULL;
 CacheRec = CGetOldest();
 DeleteSample(CacheRec);
 for (i = 0; i < DelList->Count; i++)
 {
  CacheRec = (TRecHeader *)DelList->Items[i];
  if (CacheRec->Marker == 'X') return CacheRec;
 }
 return NULL;
}
//---------------------------------------------------------------------------
TRecHeader* __fastcall TFileStorage::GetFreeBlock(int dID)
{
 TRecHeader *CacheRec, *BlockHdr = NULL;
 int i, CacheOffset = sizeof(TStorageHeader);
 if ((CacheRec = FindFreeBlock()) != NULL) return CacheRec;
 CacheRec = new TRecHeader;
 CacheRec->Size = -1;
 CacheRec->Index = -1;
 CacheRec->dID = dID;
 CacheRec->Tag = DIDs[dID]->LastTag;
 CacheRec->Marker = 'S';

 for (i = 0; i < CacheList->Count; i++ )
 {
  BlockHdr = (TRecHeader *)CacheList->Items[i];
  if (BlockHdr->Index < CacheCount)
  {
   CacheRec->CacheOffset = BlockHdr->CacheOffset;
   CacheRec->DataOffset = BlockHdr->DataOffset + BlockHdr->Index*FileHeader->RecSize;
   BlockHdr->Index++;
   return CacheRec;
  }
  if (CacheOffset < BlockHdr->DataOffset) CacheOffset = BlockHdr->DataOffset;
 }

 if (i > 0) CacheOffset+=CacheCount*FileHeader->RecSize;
 TDateTime *tTime = new TDateTime;
 BlockHdr = new TRecHeader;
 BlockHdr->CacheOffset = CacheOffset;
 BlockHdr->DataOffset = CacheOffset;
 BlockHdr->Marker = 'C';
 BlockHdr->StartDateTime = tTime->CurrentDateTime();
 BlockHdr->Index = 2;
 BlockHdr->Size = sizeof(TRecHeader);
 BlockHdr->dID = dID;
 BlockHdr->Tag = -1;
 AddCacheRec(BlockHdr, 0);
 CacheRec->CacheOffset = BlockHdr->CacheOffset;
 CacheRec->DataOffset = BlockHdr->DataOffset + FileHeader->RecSize;
 delete tTime;
 return CacheRec;
}
//---------------------------------------------------------------------------
TRecHeader* __fastcall TFileStorage::CGetOldest()
{
 double time;
 TRecHeader *CacheRec, *Result;
 int i = 0;
 if (FirstList->Count == 0) return NULL;
 CacheRec = (TRecHeader *)FirstList->Items[i];
 time = CacheRec->StartDateTime;
 Result = CacheRec;
 for (i = 1; i < FirstList->Count; i++)
 {
  CacheRec = (TRecHeader *)FirstList->Items[i];
  if (CacheRec->StartDateTime < time)
  {
   time = CacheRec->StartDateTime;
   Result = CacheRec;
  }
 }
 return  Result;
}
//---------------------------------------------------------------------------
TRecHeader* __fastcall TFileStorage::GetCacheBlock(int dID, int Tag, int Sample)
{
 TRecHeader *CacheRec, *BlockHdr = NULL;
 int i, j, CacheOffset = -1, LSample = -1, HSample = 2147483647;
 TRecHeader *LRec = new TRecHeader;
 LRec->Marker = '#';
 TRecHeader *HRec = new TRecHeader;
 HRec->Marker = '#';
 char Data[BUF_SIZE];
 for (i = 0; i < FlushList->Count; i++)
 {
  CacheRec = (TRecHeader *)FlushList->Items[i];
  if (CacheRec->Marker == 'C') continue;
  if ((CacheRec->dID != dID)||(CacheRec->Tag != Tag)) continue;
  if (CacheRec->Index == Sample) return CacheRec;
  if ((CacheRec->Marker == 'T')&&(Sample == -1)) return CacheRec;
 }
 for (i = 0; i < FirstList->Count; i++)
 {
  CacheRec = (TRecHeader *)FirstList->Items[i];
  if ((CacheRec->dID == dID)&&(CacheRec->Tag == Tag))
  {
   if (CacheRec->Index == Sample) return CacheRec;
   CacheOffset = CacheRec->CacheOffset;
   break;
  }
 }
 if (CacheOffset < 0) return NULL;
 for (i = 0; i < CacheList->Count; i++)
 {
  BlockHdr = (TRecHeader *)CacheList->Items[i];
  if (BlockHdr->DataOffset == CacheOffset) break;
 }
 if (BlockHdr == NULL) return NULL;
 j = i;
 do
 {
  DataFile->Seek(BlockHdr->CacheOffset, soFromBeginning);
  DataFile->Read(Data, FileHeader->RecSize);
  for (int ofs = sizeof(TRecHeader); ofs < BlockHdr->Size; ofs+=sizeof(TRecHeader))
  {
   CacheRec = (TRecHeader *)&Data[ofs];
   if (CacheRec->Marker == 'C') continue;
   if ((CacheRec->dID == dID) && (CacheRec->Tag == Tag))
   {
    if (CacheRec->Index == Sample)
    {
     CacheRec = new TRecHeader;
     memcpy(CacheRec, &Data[ofs], sizeof(TRecHeader));
     return CacheRec;
    }
    if (LSample < CacheRec->Index)
    {
     LSample = CacheRec->Index;
     memcpy(LRec, CacheRec, sizeof(TRecHeader));
    }
    if (HSample > CacheRec->Index)
    {
     HSample = CacheRec->Index;
     memcpy(HRec, CacheRec, sizeof(TRecHeader));
    }
   }
  }
  j++;
  if (j >= CacheList->Count) j = 0;
  BlockHdr = (TRecHeader *)CacheList->Items[j];
 } while (j!=i);
 if ((Sample < 0)&&(LRec->Marker!='#')) return LRec;
 if ((Sample == 0)&&(HRec->Marker!='#')) return HRec;
 return NULL;
}
//---------------------------------------------------------------------------
int __fastcall TFileStorage::DeleteSample(TRecHeader *CacheRec)
{
 int result = 0;
 TRecHeader *CRec;
 if (CacheRec->Marker == 'T')
 {
  for (int k = 0; k < TagList->Count; k++)
  if ((((TTagHeader *)TagList->Items[k])->dID == CacheRec->dID) &&
      (((TTagHeader *)TagList->Items[k])->Tag == CacheRec->Tag))
  {
   TagList->Delete(k);
   delete (TTagHeader *)TagList->Items[k];
  }
 }
 if (FlushList->IndexOf(CacheRec) < 0) FlushList->Add(CacheRec);
 if (DelList->IndexOf(CacheRec) < 0) DelList->Add(CacheRec);
 if (CacheRec->Marker == 'F')
 {
  result = -1;
  CRec = GetCacheBlock(CacheRec->dID, CacheRec->Tag, CacheRec->Index + 1);
  FirstList->Remove( CacheRec );
  if (CRec != NULL)
  {
   CRec->Marker = 'F';
   AddCacheRec(CRec, NOTFLUSHED);
  }
 }
 CacheRec->Marker = 'X';
 return result;
}
//---------------------------------------------------------------------------
long __fastcall TFileStorage::Offset(int Sample, int Shift)
{
 return Sample*FileHeader->RecSize + Shift;
}
//---------------------------------------------------------------------------
int __fastcall TFileStorage::Sample(long Offset)
{
 if (Offset < FileHeader->RecSize) return 0;
 int Shift = Offset % FileHeader->RecSize;
 return (Offset - Shift) / FileHeader->RecSize;
}
//---------------------------------------------------------------------------
int __fastcall TFileStorage::Shift(long Offset)
{
 if (Offset < FileHeader->RecSize) return Offset;
 return Offset % FileHeader->RecSize;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void __fastcall TFileStorage::FlushCache(int Mode)
{
 TRecHeader *CacheRec, *BlockHdr = NULL, *RecHeader;
 char Data[BUF_SIZE];
 if (FlushList->Count == 0) return;
 if ((FlushList->Count < CacheCount)&&(Mode == FM_EASY)) return;
 int i, j = 0;
 while (j < FlushList->Count)
 {
  if (FlushList->Items[j] == NULL) { j++; continue; }
  CacheRec = (TRecHeader *)FlushList->Items[j];
  for (i = 0; i < CacheList->Count; i++)
  {
   BlockHdr = (TRecHeader *)CacheList->Items[i];
   if (BlockHdr->DataOffset == CacheRec->CacheOffset) break;
  }
  DataFile->Seek(BlockHdr->DataOffset, soFromBeginning);
  DataFile->Read(Data, FileHeader->RecSize);

  for (i = j; i < FlushList->Count; i++)
  {
   if (FlushList->Items[i] == NULL) continue;
   CacheRec = (TRecHeader *)FlushList->Items[i];
   if (CacheRec->CacheOffset != BlockHdr->DataOffset) continue;
   for (int ofs = sizeof(TRecHeader); ofs < BlockHdr->Size; ofs+=sizeof(TRecHeader))
   {
    RecHeader = (TRecHeader *)&Data[ofs];
    if (RecHeader->DataOffset != CacheRec->DataOffset) continue;
    memcpy(RecHeader, CacheRec, sizeof(TRecHeader));
    FlushList->Items[i] = NULL;
    break;
   }
   if (FlushList->Items[i] != NULL)
   {
    memcpy(&Data[BlockHdr->Size], CacheRec, sizeof(TRecHeader));
    BlockHdr->Size+=sizeof(TRecHeader);
    FlushList->Items[i] = NULL;
   }
  }
  memcpy(&Data[0], BlockHdr, sizeof(TRecHeader));
  DataFile->Seek(BlockHdr->CacheOffset, soFromBeginning);
  DataFile->Write(Data, FileHeader->RecSize);
 }
 FlushList->Pack();
}
//---------------------------------------------------------------------------
void __fastcall TFileStorage::AddCacheRec(TRecHeader *CacheRec, int Mode)
{
 int i;
 if (Mode & NOTFLUSHED)
 {
  for (i = 0; i < FlushList->Count; i++)
   if (((TRecHeader *)FlushList->Items[i])->DataOffset == CacheRec->DataOffset) return;
  if (FlushList->IndexOf(CacheRec) < 0) FlushList->Add(CacheRec);
 }
 switch (CacheRec->Marker) {
  case 'C':
   if (Mode & CHECKED)
    for (i = 0; i < CacheList->Count; i++)
     if (((TRecHeader *)CacheList->Items[i])->DataOffset == CacheRec->DataOffset) return;
   if (CacheList->IndexOf(CacheRec) < 0) CacheList->Add(CacheRec);
  break;
  case 'F':
   if (Mode & CHECKED)
    for (i = 0; i < FirstList->Count; i++)
     if (((TRecHeader *)FirstList->Items[i])->DataOffset == CacheRec->DataOffset) return;
   if (FirstList->IndexOf(CacheRec) < 0) FirstList->Add(CacheRec);
  break;
  case 'X':
   if (Mode & CHECKED)
    for (i = 0; i < DelList->Count; i++)
     if (((TRecHeader *)DelList->Items[i])->DataOffset == CacheRec->DataOffset) return;
   if (DelList->IndexOf(CacheRec) < 0) DelList->Add(CacheRec);
  break;
 }
}
//---------------------------------------------------------------------------
void __fastcall TFileStorage::FreeCacheRec(TRecHeader *CacheRec)
{
 if ((FlushList->IndexOf(CacheRec) < 0) && (CacheList->IndexOf(CacheRec) < 0) &&
     (FirstList->IndexOf(CacheRec) < 0) && (DelList->IndexOf(CacheRec) < 0))
     delete CacheRec;
}
//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//
static inline void ValidCtrCheck(TDataStorage *)
{
    new TDataStorage(NULL);
}
//---------------------------------------------------------------------------
void __fastcall TDataStorage::SetConfig(AnsiString asConfig)
{
 int i, k=0;
 AnsiString asKeyWord[DATWORDS] = { "AddFile", "AddGroup" };
 AnsiString asTemp, asParam, asDev;
 FConfig = asConfig;
 if (FConfig < "  ") return;
 FConfigList->LoadFromFile(FConfig);
 for (i=0; i<FConfigList->Count; i++)
 {
  asTemp=FConfigList->Strings[i];
  asDev=asTemp.SubString(asTemp.Pos("<")+1,asTemp.Pos(">")-asTemp.Pos("<")-1);
  asParam=asTemp.SubString(asTemp.Pos("{")+1,asTemp.Pos("}")-asTemp.Pos("{")-1);
  while ((k<DATWORDS)&&(!asTemp.Pos(asKeyWord[k]))) k++;
  switch (k) {
   case 0:
    AddFile(asDev);
   break;
   case 1:
    AddDataGroup(asParam.ToInt(), asDev);
   break;
  }
 }
}
//---------------------------------------------------------------------------
__fastcall TDataStorage::TDataStorage(TComponent* Owner)
    : TComponent(Owner)
{
 int i;
 FDataFiles = new TStringList();
 for (i = 0; i < MAX_FILES; i++) DataGroup[i] = NULL;
 FConfigList = new TStringList();
 SetConfig(FConfig);
 FMinBytes = 24576;
}
//---------------------------------------------------------------------------
void __fastcall TDataStorage::Clear()
{
 for (int i=0; i<FDataFiles->Count; i++) delete FDataFiles->Objects[i];
 delete FDataFiles;
 FConfig = "";
}
//---------------------------------------------------------------------------
__fastcall TDataStorage::~TDataStorage()
{
 int i, j;
 TFileStorage *FileStorage;
 for (i = 0; i < FDataFiles->Count; i++)
 {
  FileStorage = (TFileStorage *)FDataFiles->Objects[i];
  for (j = 0; j < MAX_DID; j++)
   if (FileStorage->DIDs[j]->Marker!='X') Close(j);
  delete FileStorage;
 }
 delete FDataFiles;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::SetMaxFileSize(int dID, int MaxSize)
{
 TFileStorage *FileStorage;
 if ((FileStorage = DataGroup[dID])==NULL) return -2;
 if (FileStorage->DataSamples > MaxSize)
   return FileStorage->DataSamples;
 FileStorage->FileHeader->MaxFileSize = MaxSize;
 return MaxSize;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::AddFile(AnsiString asFileName)
{
 TFileStorage *FileStorage;
 int i;
 if ((i = FDataFiles->IndexOf(asFileName))!=-1) return i;
 FileStorage = new TFileStorage(asFileName, FOnInitCacheBlock, FOnInit);
 FDataFiles->AddObject(asFileName, FileStorage);
 return FDataFiles->Count-1;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::SetMaxFileSize(AnsiString FileName, int MaxSize)
{
 int d;
 TFileStorage *FileStorage;
 if ((d = FDataFiles->IndexOf(FileName)) < 0) return -1;
 if ((FileStorage = (TFileStorage *)FDataFiles->Objects[d])==NULL) return -2;
 if (FileStorage->DataSamples > MaxSize)
   return FileStorage->DataSamples;
 FileStorage->FileHeader->MaxFileSize = MaxSize;
 return MaxSize;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::GetMaxFileSize(AnsiString FileName)
{
 int d;
 TFileStorage *FileStorage;
 if ((d = FDataFiles->IndexOf(FileName)) < 0) return -1;
 if ((FileStorage = (TFileStorage *)FDataFiles->Objects[d])==NULL) return -2;
 return FileStorage->FileHeader->MaxFileSize;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::GetCurFileSize(AnsiString FileName)
{
 int d;
 TFileStorage *FileStorage;
 if ((d = FDataFiles->IndexOf(FileName)) < 0) return -1;
 if ((FileStorage = (TFileStorage *)FDataFiles->Objects[d])==NULL) return -2;
 return FileStorage->DataSamples;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::RemoveFile(int fHandle)
{
 if (fHandle>=FDataFiles->Count) return -1;
 delete FDataFiles->Objects[fHandle];
 DeleteFile(FDataFiles->Strings[fHandle]);
 FDataFiles->Strings[fHandle] = "!@#$%DELETED";
 return 0;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::AddDataGroup(int dID, int fHandle)
{
 if (dID>MAX_FILES) return -1;
 if (DataGroup[dID]!=NULL) return -2;
 if (fHandle>=FDataFiles->Count) return -3;
 TFileStorage *FileStorage = (TFileStorage*)FDataFiles->Objects[fHandle];
 DataGroup[dID]=FileStorage;
 return 0;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::AddDataGroup(int dID, AnsiString asFileName)
{
 if (dID>MAX_FILES) return -1;
 if (DataGroup[dID]!=NULL) return -2;
 for (int i = 0; i<FDataFiles->Count; i++)
  if (FDataFiles->Strings[i]==asFileName)
  {
   TFileStorage *FileStorage = (TFileStorage*)FDataFiles->Objects[i];
   FileStorage->DIDs[dID]->LastTag = -1;
   DataGroup[dID]=FileStorage;
   return 0;
  }
 return -1;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::RemoveDataGroup(int dID)
{
 DataGroup[dID] = NULL;
 return 0;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::GetStorage(int dID)
{
 if (DataGroup[dID]==NULL) return -1;
 return FDataFiles->IndexOfObject(DataGroup[dID]);
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::MoveDataGroup(int dID, int SfHandle, int DfHandle)
{
 if (dID>MAX_FILES) return -1;
 if (DataGroup[dID]!=NULL) return -2;
 if (DfHandle>=FDataFiles->Count) return -3;
 DataGroup[dID]=(TFileStorage *)FDataFiles->Objects[DfHandle];
 return 0;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::Read(int dID, int Tag, int Sample, char *ptr)
{
 TFileStorage *FileStorage;
 TRecHeader *CacheRec;
 int Readed;
 if (dID>MAX_FILES) return -1;
 if ((FileStorage = DataGroup[dID])==NULL) return -2;
 if ((Sample == FileStorage->DIDs[dID]->LastIndex)&&(Tag == FileStorage->DIDs[dID]->LastTag))
 {
  memcpy(ptr, FileStorage->DIDs[dID]->Data, FileStorage->DIDs[dID]->BufSize);
  return FileStorage->DIDs[dID]->BufSize;
 }
 if ((CacheRec = FileStorage->GetCacheBlock(dID, Tag, Sample)) == NULL) return -1;
 FileStorage->DataFile->Seek(CacheRec->DataOffset, soFromBeginning);
 if (CacheRec->Marker == 'T') FileStorage->DataFile->Seek(sizeof(TTagHeader), soFromCurrent);
 Readed = FileStorage->DataFile->Read(ptr, CacheRec->Size);
 FileStorage->FreeCacheRec(CacheRec);
 return Readed;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::SRead(int Handle, char *ptr, long cnt)
{
 int Sample, Shift;
 int dID    = Handles[Handle]->dID;
 int Tag    = Handles[Handle]->Tag;
 if (( dID < 0 ) || ( Tag < 0 )) return -1;
 int RD1, RD2, RD3 = cnt;
// char *buf = (char *)malloc(RecSize);
 char buf[BUF_SIZE];

 while (1)
 {
  Sample = DataGroup[dID]->Sample(Handles[Handle]->Offset);
  Shift = DataGroup[dID]->Shift(Handles[Handle]->Offset);
  RD1 = Read(dID, Tag, Sample, buf);
  if (RD1 < 0) return RD1;
  if ((RD1-=Shift) == 0) return cnt - RD3;
  if (RD3 > RD1) RD2 = RD1;
  else RD2 = RD3;
  memcpy(ptr + (cnt - RD3), buf + Shift, RD2);
  RD3-=RD2;
  Handles[Handle]->Offset+=RD2;
  if ( RD3 <= 0 )
  {
//   SayToLog(dID, "SRead" + AnsiString(cnt) + ", readed:" + AnsiString(cnt - RD3));
   return cnt - RD3;
  }
 }
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::SWrite(int Handle, char *ptr, long cnt)
{
 int dID    = Handles[Handle]->dID;
 if ( dID < 0 )
 {
  SayToLog(dID, "SWrite error -1");
  return -1;
 }
 int iR = Write(dID, ptr, cnt);
 if ( iR < 0 ) SayToLog(dID, "SWrite error " + AnsiString(iR));
 else
 {
//  SayToLog(dID, "SWrite Ok " + AnsiString(iR));
  Handles[Handle]->Offset+=iR;
 }
 return iR;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::SSeek(int Handle, long offset, int fromwhere)
{
 long new_offset, Sample;
 int dID = Handles[Handle]->dID;
 int RecSize   = DataGroup[dID]->FileHeader->RecSize;
// AnsiString    aFW;
 TFileStorage *FileStorage;
 FileStorage = DataGroup[dID];
 switch (fromwhere) {
  case SEEK_SET:
   new_offset = offset;
//   aFW = "SEEK_SET";
  break;
  case SEEK_CUR:
   new_offset = Handles[Handle]->Offset + offset;
//   aFW = "SEEK_CUR";
  break;
  case SEEK_END:
   new_offset = FileStorage->DIDs[dID]->Bytes + offset;
//   aFW = "SEEK_END";
  break;
 }
 Sample = new_offset / RecSize;
 if ( new_offset > FileStorage->DIDs[dID]->Bytes )
  SayToLog(dID, "SSeek error: Bytes < new_offset");

 Handles[Handle]->Offset = new_offset;
// SayToLog(dID, "SSeek Ok " + AnsiString(Handles[Handle]->Offset) + ", " + aFW);
 SetBuffer(dID, Sample);
 FileStorage->DIDs[dID]->BufSize = FileStorage->Shift(new_offset);
 return Handles[Handle]->Offset;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::SetBuffer(int dID, int Sample)
{
 int Tag, HS = 0;
 TFileStorage *FileStorage;
 if ((FileStorage = DataGroup[dID]) == NULL) return -1;
 if ((Tag = FileStorage->DIDs[dID]->LastTag) < 0) return -1;
 TRecHeader *CacheRec;
 if (Sample == FileStorage->DIDs[dID]->LastIndex) return 0;
 FlushBuffer(dID);
 if ((CacheRec = FileStorage->GetCacheBlock(dID, Tag, Sample)) == NULL) return -1;
 FileStorage->AddCacheRec(CacheRec, NOTFLUSHED);
 FileStorage->DIDs[dID]->LastIndex = Sample;
 FileStorage->DIDs[dID]->Marker = CacheRec->Marker;
 if (CacheRec->Marker == 'T') HS = sizeof(TTagHeader);
 FileStorage->DataFile->Seek(CacheRec->DataOffset + HS, soFromBeginning);
 FileStorage->DataFile->Read(FileStorage->DIDs[dID]->Data, DataGroup[dID]->FileHeader->RecSize);
 return 0;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::FlushBuffer(int dID)
{
 int Written, HS = 0;
 TFileStorage *FileStorage;
 TRecHeader *CacheRec;
 FileStorage = DataGroup[dID];
 if (FileStorage->DIDs[dID]->LastTag < 0) return 0;
 if ((CacheRec = FileStorage->GetCacheBlock(dID, FileStorage->DIDs[dID]->LastTag,
                                        FileStorage->DIDs[dID]->LastIndex)) == NULL)
                                            CacheRec = FileStorage->GetFreeBlock(dID);

 if (CacheRec->Size < 0) FileStorage->DataSamples++;
 CacheRec->dID = dID;
 CacheRec->Tag = FileStorage->DIDs[dID]->LastTag;
 CacheRec->Index = FileStorage->DIDs[dID]->LastIndex;
 CacheRec->Marker = FileStorage->DIDs[dID]->Marker;
 if (CacheRec->Index != FileStorage->Sample(FileStorage->DIDs[dID]->Bytes))
  CacheRec->Size = FileStorage->FileHeader->RecSize;
 else
  CacheRec->Size = FileStorage->Shift(FileStorage->DIDs[dID]->Bytes);

 FileStorage->DataFile->Seek(CacheRec->DataOffset, soFromBeginning);
 if (FileStorage->DIDs[dID]->Marker == 'T')
 {
  HS = sizeof(TTagHeader);
  FileStorage->DIDs[dID]->Header->TagSize = FileStorage->DIDs[dID]->Bytes;
  FileStorage->DataFile->Write(FileStorage->DIDs[dID]->Header, HS);
  FileStorage->DIDs[dID]->Marker = 'X';
 }
 else FileStorage->DIDs[dID]->Marker = 'S';
 Written = FileStorage->DataFile->Write(FileStorage->DIDs[dID]->Data, FileStorage->FileHeader->RecSize - HS);

 FileStorage->AddCacheRec(CacheRec, NOTFLUSHED);
 FileStorage->DIDs[dID]->LastIndex++;
 FileStorage->DIDs[dID]->BufSize = 0;
 return Written;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::Write(int dID, char *ptr, long cnt)
{
 TFileStorage *FileStorage;
 int WrSize = 0, Offset;
 if (dID>MAX_FILES) return -1;
 if ((FileStorage = DataGroup[dID])==NULL) return -2;
 if (FileStorage->DIDs[dID]->Marker == 'X') return -3;

 int BufSize = FileStorage->DIDs[dID]->BufSize;
 int BytesLeft = FileStorage->FileHeader->RecSize - BufSize;

 while (cnt - WrSize >= BytesLeft)
 {
  memcpy(&FileStorage->DIDs[dID]->Data[BufSize], ptr + WrSize, BytesLeft);
  FileStorage->DIDs[dID]->BufSize = FileStorage->FileHeader->RecSize;
  Offset = FileStorage->Offset(FileStorage->DIDs[dID]->LastIndex, FileStorage->DIDs[dID]->BufSize);
  if (Offset > FileStorage->DIDs[dID]->Bytes) FileStorage->DIDs[dID]->Bytes = Offset;

  FlushBuffer(dID);
  FileStorage->FlushCache(FM_EASY);
  WrSize+=BytesLeft;
  BufSize = FileStorage->DIDs[dID]->BufSize;
  BytesLeft = FileStorage->FileHeader->RecSize - BufSize;
 }

 BytesLeft = FileStorage->FileHeader->RecSize - FileStorage->DIDs[dID]->BufSize;
 if ((WrSize < cnt)&&(cnt - WrSize <= BytesLeft))
 {
  memcpy(&FileStorage->DIDs[dID]->Data[BufSize], ptr + WrSize, cnt - WrSize);
  FileStorage->DIDs[dID]->BufSize+= cnt - WrSize;
  WrSize = cnt;
 }
 Offset = FileStorage->Offset(FileStorage->DIDs[dID]->LastIndex, FileStorage->DIDs[dID]->BufSize);
 if (Offset > FileStorage->DIDs[dID]->Bytes) FileStorage->DIDs[dID]->Bytes = Offset;
 return WrSize;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::Delete(int dID, int Tag)
{
 int Sample = 0;
 TFileStorage *FileStorage;
 TRecHeader *CRec;
 if ((FileStorage = DataGroup[dID])==NULL) return -2;
 while ((CRec = FileStorage->GetCacheBlock(dID, Tag, Sample)) == NULL) Sample++;
 do
 {
  FileStorage->DeleteSample(CRec);
  CRec = FileStorage->GetCacheBlock(dID, Tag, Sample);
  Sample++;
 }
 while ( CRec != NULL );
 return 0;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::Close(int dID)
{
 TFileStorage *FileStorage;
 TDateTime *CTime = new TDateTime();
 *CTime = CTime->CurrentDateTime();
 if (dID>MAX_FILES) return -1;
 if ((FileStorage = DataGroup[dID])==NULL) return -2;
 if (FileStorage->DIDs[dID]->LastIndex != FileStorage->Sample(FileStorage->DIDs[dID]->Bytes))
  SetBuffer(dID, FileStorage->Sample(FileStorage->DIDs[dID]->Bytes));


 if (BUF_SIZE - FileStorage->DIDs[dID]->BufSize < sizeof(TTagHeader)) FlushBuffer(dID);
 FileStorage->DIDs[dID]->BufSize = FileStorage->Shift(FileStorage->DIDs[dID]->Bytes);
 FileStorage->DIDs[dID]->Marker = 'T';
 FileStorage->DIDs[dID]->Header->LengthTime = *CTime - FileStorage->DIDs[dID]->Header->StartDateTime;
 FlushBuffer(dID);
 FileStorage->DIDs[dID]->LastTag = -1;
 FileStorage->DIDs[dID]->Marker = 'X';
 FileStorage->DIDs[dID]->Header = NULL;
 delete CTime;
 FileStorage->FlushCache(FM_EASY);
 return 0;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::Open(int dID)
{
 TFileStorage *FileStorage;
 TDateTime *DateT = new TDateTime();
 *DateT = DateT->CurrentDateTime();
 if (dID>MAX_FILES) return -1;
 if ((FileStorage = DataGroup[dID])==NULL) return -2;
 FileStorage->FileHeader->LastTag[dID]++;
 FileStorage->DIDs[dID]->LastTag = FileStorage->FileHeader->LastTag[dID];
 FileStorage->DIDs[dID]->Marker = 'F';
 FileStorage->DIDs[dID]->Bytes = 0;
 FileStorage->DIDs[dID]->BufSize = 0;
 FileStorage->DIDs[dID]->LastIndex = 0;

 FileStorage->DIDs[dID]->Header = new TTagHeader;
 FileStorage->DIDs[dID]->Header->StartDateTime = *DateT;
 FileStorage->DIDs[dID]->Header->TagSize = -1;
 FileStorage->DIDs[dID]->Header->LengthTime = -1;
 FileStorage->DIDs[dID]->Header->Reads = 0;
 FileStorage->DIDs[dID]->Header->Key8[0] = 0;
 FileStorage->DIDs[dID]->Header->Key32[0] = 0;
 FileStorage->DIDs[dID]->Header->Compressor = -1;
 FileStorage->DIDs[dID]->Header->dID = dID;
 FileStorage->DIDs[dID]->Header->Tag = FileStorage->DIDs[dID]->LastTag;
 FileStorage->TagList->Add(FileStorage->DIDs[dID]->Header);
 delete DateT;
 return FileStorage->DIDs[dID]->LastTag;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::SOpen(int dID, int Tag)
{
 SayToLog(dID, "SOpen");
 for ( int i = 0; i < MAX_HANDLES; i++ )
 {
  if (Handles[i] == NULL)
  {
   Handles[i] = new TSHandle();
   Handles[i]->dID = -1;
  }
  if (Handles[i]->dID < 0)
  {
   if (Tag < 0)
   {
    Handles[i]->Mode = fmCreate;
    if ((Handles[i]->Tag = Open(dID)) < 0 ) return Handles[i]->Tag;
   }
   else
   {
    Handles[i]->Tag = Tag;
    Handles[i]->Mode = fmOpenRead;
   }
   Handles[i]->dID = dID;
   Handles[i]->Offset = 0;
   return i;
  }
 }
 return -1;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::SClose(int Handle)
{
 int iR;
 if (Handles[Handle]->Mode == fmCreate)
 {
//  SSeek(Handle, 0, SEEK_END);
  if ((iR = Close(Handles[Handle]->dID)) < 0) return iR;
 }
 SayToLog(Handles[Handle]->dID, "SClose " + AnsiString(iR));
 Handles[Handle]->dID = -1;
 Handles[Handle]->Tag = -1;
 return 0;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::H2dID(int Handle)
{
 return Handles[Handle]->dID;
}
int __fastcall TDataStorage::H2Tag(int Handle)
{
 return Handles[Handle]->Tag;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::InsertFile(int dID, System::AnsiString FileName)
{
 char buf[BUF_SIZE];
 int Written = 0, Readed, BytesReaded = 0;
 if (Open(dID) < 0) return -1;
 int InsertFile = FileOpen(FileName, fmOpenRead);
 int iFileLength = FileSeek(InsertFile,0,2);
 FileSeek(InsertFile,0,0);
 while (BytesReaded < iFileLength)
 {
  Readed = FileRead(InsertFile, buf, BUF_SIZE);
  Written+=Write(dID, buf, Readed);
  BytesReaded+=Readed;
 }
 if (Close(dID) < 0) return -1;
 FileClose(InsertFile);
 return Written;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::SaveToFile(int dID, int Tag, System::AnsiString FileName)
{
 int Sample = 0;
 int Readed;
 int Size=0;
 char Data[BUF_SIZE];
 TFileStream *SaveFile = new TFileStream(FileName, fmCreate|fmShareExclusive);
 do
 {
  Readed = Read(dID, Tag, Sample, Data);
  if (Readed > 0) SaveFile->Write(Data, Readed);
  Size+=Readed;
  Sample++;
 } while (Readed>0);
 delete SaveFile;
 return Size;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::GetFileSize(int fHandle)
{
 TFileStorage *FileStorage = (TFileStorage *)FDataFiles->Objects[fHandle];
 return FileStorage->DataFile->Size / 1024;
}
//---------------------------------------------------------------------------
TList* __fastcall TDataStorage::GetTagList(int fHandle)
{
 TFileStorage *FileStorage;
 if ((FileStorage = (TFileStorage *)FDataFiles->Objects[fHandle]) == NULL) return NULL;
 return FileStorage->TagList;
}
//---------------------------------------------------------------------------
void __fastcall TDataStorage::Map(TStrings *IDList, int fHandle)
{
 TFileStorage *FileStorage;
 char Data[BUF_SIZE];
 TTagHeader *TagHeader = new TTagHeader;
 TRecHeader *RecHeader, *BlockHdr;
 TDateTime *tTime, *lTime;
 FileStorage = (TFileStorage *)FDataFiles->Objects[fHandle];

 FileStorage->DataFile->Seek(sizeof(TStorageHeader), soFromBeginning);
 while (1)
 {
  FileStorage->DataFile->Read(Data, BUF_SIZE);
  BlockHdr = (TRecHeader *)&Data[0];
  for (int i = 0; i < BlockHdr->Index; i++)
  {
   RecHeader = (TRecHeader *)&Data[i*sizeof(TRecHeader)];
   if (RecHeader->Marker=='T')
   {
    FileStorage->DataFile->Seek(RecHeader->DataOffset, soFromBeginning);
    FileStorage->DataFile->Read(TagHeader, sizeof(TTagHeader));
    IDList->Add("-----------TagHeader-----------");
   }
   IDList->Add(AnsiString(RecHeader->dID)+", "+ AnsiString(RecHeader->Tag) +
              ", Index:"+AnsiString(RecHeader->Index) +
              " RecSize: "+AnsiString(RecHeader->Size)+", "+ AnsiString(RecHeader->Marker));
   if (RecHeader->Marker=='T')
   {
    tTime = new TDateTime(TagHeader->StartDateTime);
    lTime = new TDateTime(TagHeader->LengthTime);
    IDList->Add("Time: "+tTime->FormatString("dd/mmm/yy hh:mm:ss")+
                       " Len: " + lTime->FormatString("hh:mm:ss")+
                       " Size: "+ AnsiString(TagHeader->TagSize));
    delete tTime;
    delete lTime;
   }
  }
  FileStorage->DataFile->Seek(BlockHdr->CacheOffset+BlockHdr->Index*FileStorage->FileHeader->RecSize, soFromBeginning);
  if (FileStorage->DataFile->Position >= FileStorage->DataFile->Size) break;
 }
 delete TagHeader;
}
//---------------------------------------------------------------------------
void __fastcall TDataStorage::List(TStrings *IDList, int fHandle)
{
 TTagHeader *TagHeader;
 TFileStorage *FileStorage;
 TDateTime *tTime, *lTime;
 FileStorage = (TFileStorage *)FDataFiles->Objects[fHandle];

 for (int i = 0; i < FileStorage->TagList->Count;i++ )
 {
  TagHeader = (TTagHeader *)FileStorage->TagList->Items[i];
  tTime = new TDateTime(TagHeader->StartDateTime);
  lTime = new TDateTime(TagHeader->LengthTime);
  IDList->Add("dID: "+AnsiString(TagHeader->dID)+" Tag: "+AnsiString(TagHeader->Tag)+
              "Time: "+tTime->FormatString("dd/mmm/yy hh:mm:ss")+
              " Len: " + lTime->FormatString("hh:mm:ss")+
              " Size: "+ AnsiString(TagHeader->TagSize)+
              " Key32: "+ AnsiString(TagHeader->Key32));
  delete tTime;
  delete lTime;
 }
}
//---------------------------------------------------------------------------
AnsiString __fastcall TDataStorage::GetCKey(int dID, int Tag, int KeyTag)
{
 TTagHeader *TagHeader = NULL;
 TFileStorage *FileStorage;
 if ((FileStorage = DataGroup[dID]) == NULL) return NULL;
 for (int i = 0; i < FileStorage->TagList->Count; i++)
 {
  TagHeader = (TTagHeader *)FileStorage->TagList->Items[i];
  if ((TagHeader->dID == dID)&&(TagHeader->Tag == Tag)) break;
 }
 if (TagHeader == NULL) return NULL;
 switch (KeyTag) {
  case KT_8 : return AnsiString(TagHeader->Key8);
  case KT_32: return AnsiString(TagHeader->Key32);
 }
 return NULL;
}
//---------------------------------------------------------------------------
int __fastcall TDataStorage::SetCKey(int dID, int Tag, int KeyTag, AnsiString asKey)
{
 AnsiString Result;
 TTagHeader *TagHeader;
 TRecHeader *CacheBlock;
 TFileStorage *FileStorage;
 if ((FileStorage = DataGroup[dID]) == NULL) return -1;
 if ((CacheBlock = FileStorage->GetCacheBlock(dID, Tag, -1)) == NULL) return -1;
 for (int i = 0; i < FileStorage->TagList->Count; i++)
 {
  TagHeader = (TTagHeader *)FileStorage->TagList->Items[i];
  if ((TagHeader->dID == dID)&&(TagHeader->Tag == Tag)) break;
 }
 if (TagHeader == NULL) return NULL;

 int aL = asKey.Length();
 switch (KeyTag) {
  case KT_8 :
   if (aL > 8) return -1;
   memcpy(TagHeader->Key8, asKey.c_str(), aL+1);
  break;
  case KT_32:
   if (aL > 32) return -1;
   memcpy(TagHeader->Key32, asKey.c_str(), aL+1);
  break;
  default : return -1;
 }
 FileStorage->DataFile->Seek(CacheBlock->DataOffset, soFromBeginning);
 FileStorage->DataFile->Write(TagHeader, sizeof(TTagHeader));
 FileStorage->FreeCacheRec(CacheBlock);
 return 0;
}
//---------------------------------------------------------------------------
void __fastcall TDataStorage::SayToLog(int dID, System::AnsiString LogPhrase)
{
 AnsiString aTime;
 AnsiString LogResult;
 TDateTime *DTime = new TDateTime();
 *DTime = DTime->CurrentDateTime();
 aTime = DTime->FormatString("dd.mm.yy:hh:mm:ss");

 LogResult = AnsiString(dID)+"|"+ aTime +" "+LogPhrase;
 if ( FLogList ) FLogList->Lines->Add(LogResult);
}
//---------------------------------------------------------------------------
namespace Datastorage
{
    void __fastcall PACKAGE Register()
    {
        TComponentClass classes[1] = {__classid(TDataStorage)};
        RegisterComponents("Additional", classes, 0);
    }
}
//---------------------------------------------------------------------------

