//---------------------------------------------------------------------------
#ifndef DataStorageH
#define DataStorageH
//---------------------------------------------------------------------------
#include <SysUtils.hpp>
#include <Controls.hpp>
#include <Classes.hpp>
#include <Forms.hpp>

#define BAT_SIZE    8192
#define BUF_SIZE    16384
#define MAX_HANDLES 1024
#define MAX_FILES   512
#define MAX_TAG     65535
#define MAX_DID     16
#define DATWORDS    2

#define FM_EASY     0
#define FM_FULL     1

#define NOTFLUSHED  1
#define CHECKED     2


#define KT_8        0
#define KT_32       1
//---------------------------------------------------------------------------
struct TStorageHeader
{
       int  DVer;
       int  RecSize;
       int  MaxFileSize;
       int  LastTag[MAX_DID];
};
//---------------------------------------------------------------------------
struct TRecHeader
{
       int  dID;
       int  Tag;
       int  Size;   // Size of  data + TagHeader
       int  Index;
       int  DataOffset; // Physycal data position in DataFile
       int  CacheOffset; // Physycal Chache Block position in DataFile
       char Marker;
       double  StartDateTime;
};
//---------------------------------------------------------------------------
struct TTagHeader
{
       int  dID;
       int  Tag;
       int  TagSize; // Full size of data in tag
       double  StartDateTime;
       double  LengthTime;
       int  Reads;
       char Compressor;
       char Key8[8];
       char Key32[32];
};
//---------------------------------------------------------------------------
struct TDIDState
{
 int  LastTag;
 int  LastIndex;
 int  Bytes;
 int  BufSize;
 char Marker;
 TTagHeader *Header;
 char Data[BUF_SIZE];
};
//---------------------------------------------------------------------------
typedef void __fastcall ( __closure *TFOnInit)(int EstimatedBlocks);
//---------------------------------------------------------------------------
class TFileStorage: public TObject
{
private:
       TRecHeader* __fastcall GetFreeCacheBlock(void);
       TRecHeader* __fastcall FindFreeBlock();
       int CacheCount;
       TList *FlushList; // List of not flushe RecHeaders
       TList *CacheList;  // List of all RecHeaders with C marker
       TList *DelList;    // List of all RecHeaders with X marker
       TList *FirstList;  // Lisr of all RecHeaders with F marker

public:
       __fastcall TFileStorage(System::AnsiString asFile, TNotifyEvent FOnInitCacheBlock,
                                                         TFOnInit FOnInit);
       __fastcall ~TFileStorage();
       void __fastcall AddCacheRec(TRecHeader *CacheRec, int Mode);
       void __fastcall FreeCacheRec(TRecHeader *CacheRec);
       void __fastcall FlushCache(int Mode);
       TRecHeader* __fastcall GetFreeBlock(int dID);
       TRecHeader* __fastcall GetCacheBlock(int dID, int Tag, int Sample);
       TRecHeader* __fastcall CGetOldest();
       int __fastcall DeleteSample(TRecHeader *CacheRec);
       long __fastcall Offset(int Sample, int Shift);
       int __fastcall Sample(long Offset);
       int __fastcall Shift(long Offset);
       TList *TagList; // List of all TagHeaders
       TStorageHeader *FileHeader;
       TFileStream *DataFile;
       TDIDState *DIDs[MAX_DID];
       int  DataSamples;
};
//---------------------------------------------------------------------------
class TSHandle
{
public:
    int dID;
    int Tag;
    int Offset;
    int Mode;
};
//---------------------------------------------------------------------------
class PACKAGE TDataStorage : public TComponent
{
private:
    TStringList *FDataFiles;
    TFileStorage *DataGroup[MAX_FILES];
    TStringList *FConfigList;
    AnsiString FConfig;
    int __fastcall FlushBuffer(int dID);
    int __fastcall SetBuffer(int dID, int Sample);
    TList* __fastcall GetTagList(int fHandle);
    TMemo* FLogList;
    TSHandle *Handles[MAX_HANDLES];
    long     FMinBytes;
    TNotifyEvent FOnInitCacheBlock;
    TFOnInit FOnInit;
protected:
public:
    void __fastcall SayToLog(int dID, System::AnsiString LogPhrase);
    __fastcall TDataStorage(TComponent* Owner);
    __fastcall ~TDataStorage();
    void __fastcall SetConfig(AnsiString asConfig);

    int __fastcall AddFile(AnsiString asFileName); // Returns internal handle for data file
    int __fastcall SetMaxFileSize(int dID, int MaxSize);
    int __fastcall SetMaxFileSize(AnsiString FileName, int MaxSize);
    int __fastcall GetMaxFileSize(AnsiString FileName);
    int __fastcall GetCurFileSize(AnsiString FileName);
    int __fastcall RemoveFile(int fHandle);
//    int __fastcall RemoveFile(AnsiString &asFileName);
    int __fastcall AddDataGroup(int dID, int fHandle);
    int __fastcall AddDataGroup(int dID, AnsiString asFileName);
    int __fastcall RemoveDataGroup(int dID);
    int __fastcall GetStorage(int dID);  // Returns assigned file handle
    int __fastcall MoveDataGroup(int dID, int SfHandle, int DfHandle);
    void __fastcall Clear();

/**/int __fastcall SOpen(int dID, int Tag); // returns virtual handle for dID don't requires Open
//  if Tag = -1 then SOpen creates new Tag for given dID
/**/int __fastcall SClose(int Handle);  // dID is NOT the same as Handle !!!!!!!!!!
/**/int __fastcall SRead(int Handle, char *ptr, long cnt);
/**/int __fastcall SWrite(int Handle, char *ptr, long cnt);
/**/int __fastcall SSeek(int Handle, long offset, int fromwhere);
/**/int __fastcall H2dID(int Handle);
/**/int __fastcall H2Tag(int Handle);
    int __fastcall Open(int dID); // Creates new Tag for current DataID and returns it
    int __fastcall Close(int dID); // Closes DataID
    int __fastcall Write(int dID, char *ptr, long cnt);
    int __fastcall Read(int dID, int Tag, int Sample, char *ptr);
    int __fastcall Delete(int dID, int Tag);      // Delete calls DeleteSample for actual del

    int __fastcall SaveToFile(int dID, int Tag, System::AnsiString FileName);
    int __fastcall InsertFile(int dID, System::AnsiString FileName);
    int __fastcall GetFileSize(int fHandle);
    AnsiString __fastcall DInfo(int dID);
    void __fastcall Map(TStrings *IDList, int fHandle);
    void __fastcall List(TStrings *IDList, int fHandle);

    AnsiString __fastcall GetCKey(int dID, int Tag, int KeyTag);
    int __fastcall SetCKey(int dID, int Tag, int KeyTag, AnsiString asKey);

    __property TList* TagList[int fHandle] = { read = GetTagList };
__published:
    __property TStringList *DataFiles = { read = FDataFiles, write = FDataFiles, nodefault };
    __property AnsiString ConfigFile = { read = FConfig, write = SetConfig };
    __property TMemo* Log = { read = FLogList, write = FLogList };
    __property long MinBytes = { read = FMinBytes, write = FMinBytes, default = 24576 };

    __property TNotifyEvent OnInitCacheBlock = { read = FOnInitCacheBlock, write = FOnInitCacheBlock};
    __property TFOnInit OnInit = { read = FOnInit, write = FOnInit};

};
//---------------------------------------------------------------------------
#endif
